use crate::ast::*;

pub enum ChildRef<'a> {
    Identifier(&'a Identifier),
    FullyQualifiedName(&'a FullyQualifiedName),
    TypeUse(&'a TypeUse),
    PrimitiveTypeUse(&'a PrimitiveTypeUse),
    InterfaceOrClassTypeUse(&'a InterfaceOrClassTypeUse),
    TupleTypeUse(&'a TupleTypeUse),
    FunctionTypeUse(&'a FunctionTypeUse),
    GenericArguments(&'a GenericArguments),
    GenericParameters(&'a GenericParameters),
    TupleArguments(&'a TupleArguments),
    ImplementsList(&'a ImplementsList),
    Parameters(&'a Parameters),
    Parameter(&'a Parameter),
    ReturnType(&'a ReturnType),
    References(&'a References),
    UseStatement(&'a UseStatement),
    ModuleLevelDeclaration(&'a ModuleLevelDeclaration),
    InterfaceLevelDeclaration(&'a InterfaceLevelDeclaration),
    ClassLevelDeclaration(&'a ClassLevelDeclaration),
    FunctionBody(&'a FunctionBody),
    ClassConstructor(&'a ClassConstructor),
    ClassConstructorParameter(&'a ClassConstructorParameter),
    BlockStatement(&'a BlockStatement),
    Statement(&'a Statement),
    IfStatement(&'a IfStatement),
    ElseIfs(&'a ElseIfs),
    ElseBlock(&'a ElseBlock),
    Expression(&'a Expression),
    TurboFish(&'a TurboFish),
    CallArguments(&'a CallArguments),
    CallArgument(&'a CallArgument),
    ClosureCaptures(&'a ClosureCaptures),
    ClosureCapture(&'a ClosureCapture),
    ClosureParameters(&'a ClosureParameters),
    ClosureParameter(&'a ClosureParameter),
    ObjectNavigations(&'a ObjectNavigations),
    ObjectNavigation(&'a ObjectNavigation),
    DString(&'a DString),
    DStringPart(&'a DStringPart),
}

pub trait HasChildren {
    fn children(&self) -> Vec<ChildRef>;
}

impl HasChildren for FullyQualifiedName {
    fn children(&self) -> Vec<ChildRef> {
        self.identifiers()
            .iter()
            .map(|id| ChildRef::Identifier(*id))
            .collect()
    }
}

impl HasChildren for TypeUse {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            TypeUse::Primitive(primitive_type_use) => primitive_type_use.children(),
            TypeUse::InterfaceOrClass(interface_or_class_type_use) => {
                interface_or_class_type_use.children()
            }
            TypeUse::Tuple(tuple_type_use) => tuple_type_use.children(),
            TypeUse::Function(function_type_use) => function_type_use.children(),
        }
    }
}

impl HasChildren for PrimitiveTypeUse {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            PrimitiveTypeUse::Array(generics) => {
                if let Some(generics) = generics {
                    vec![ChildRef::GenericArguments(&generics)]
                } else {
                    vec![]
                }
            }
            _ => vec![],
        }
    }
}

impl HasChildren for InterfaceOrClassTypeUse {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::FullyQualifiedName(self.fqn()));
        children.push(ChildRef::GenericArguments(self.generics()));
        children
    }
}

impl HasChildren for TupleTypeUse {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::TupleArguments(self.arguments())]
    }
}

impl HasChildren for FunctionTypeUse {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
        ]
    }
}

impl HasChildren for GenericArguments {
    fn children(&self) -> Vec<ChildRef> {
        self.arguments()
            .iter()
            .map(|type_use| ChildRef::TypeUse(*type_use))
            .collect()
    }
}

impl HasChildren for GenericParameters {
    fn children(&self) -> Vec<ChildRef> {
        self.identifiers()
            .iter()
            .map(|id| ChildRef::Identifier(*id))
            .collect()
    }
}

impl HasChildren for TupleArguments {
    fn children(&self) -> Vec<ChildRef> {
        self.type_uses()
            .iter()
            .map(|type_use| ChildRef::TypeUse(*type_use))
            .collect()
    }
}

impl HasChildren for ImplementsList {
    fn children(&self) -> Vec<ChildRef> {
        self.type_uses()
            .iter()
            .map(|type_use| ChildRef::TypeUse(*type_use))
            .collect()
    }
}

impl HasChildren for Parameters {
    fn children(&self) -> Vec<ChildRef> {
        self.parameters()
            .iter()
            .map(|parameter| ChildRef::Parameter(*parameter))
            .collect()
    }
}

impl HasChildren for Parameter {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Identifier(self.identifier()),
            ChildRef::TypeUse(self.type_use()),
        ]
    }
}

impl HasChildren for ReturnType {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::TypeUse(self.declared_type()),
            ChildRef::References(self.references()),
        ]
    }
}

impl HasChildren for CompilationUnit {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        if let Some(namespace) = self.namespace() {
            children.push(ChildRef::FullyQualifiedName(namespace));
        }
        children.extend(
            self.use_statements()
                .iter()
                .map(|use_statement| ChildRef::UseStatement(*use_statement)),
        );
        children.extend(
            self.declarations()
                .iter()
                .map(|declaration| ChildRef::ModuleLevelDeclaration(*declaration)),
        );
        children
    }
}

impl HasChildren for UseStatement {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.extend(
            self.identifiers()
                .iter()
                .map(|identifier| ChildRef::Identifier(*identifier)),
        );
        match self.last() {
            UseStatementLast::Identifier(identifier) => {
                children.push(ChildRef::Identifier(identifier))
            }
            UseStatementLast::Identifiers(identifiers) => children.extend(
                identifiers
                    .iter()
                    .map(|identifier| ChildRef::Identifier(identifier.clone())),
            ),
            UseStatementLast::Star => {}
        }
        children
    }
}

impl HasChildren for ModuleLevelDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        use ModuleLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.children(),
            Interface(interface_declaration) => interface_declaration.children(),
            Class(class_declaration) => class_declaration.children(),
            Function(function_definition) => function_definition.children(),
            PlatformFunction(platform_function_declaration) => {
                platform_function_declaration.children()
            }
        }
    }
}

impl HasChildren for InterfaceLevelDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        use InterfaceLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.children(),
            Interface(interface_declaration) => interface_declaration.children(),
            Class(class_declaration) => class_declaration.children(),
            Function(function_declaration) => function_declaration.children(),
            OperatorFunction(operator_function_declaration) => {
                operator_function_declaration.children()
            }
        }
    }
}

impl HasChildren for ClassLevelDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        use ClassLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.children(),
            Interface(interface_declaration) => interface_declaration.children(),
            Class(class_declaration) => class_declaration.children(),
            Function(function_definition) => function_definition.children(),
            OperatorFunction(operator_function_definition) => {
                operator_function_definition.children()
            }
            PlatformFunction(platform_function_declaration) => {
                platform_function_declaration.children()
            }
            Property(property_declaration) => property_declaration.children(),
            Field(field_declaration) => field_declaration.children(),
        }
    }
}

impl HasChildren for ModuleDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Identifier(self.identifier()));
        children.extend(
            self.declarations()
                .iter()
                .map(|declaration| ChildRef::ModuleLevelDeclaration(*declaration)),
        );
        children
    }
}

impl HasChildren for InterfaceDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Identifier(self.identifier()));
        children.push(ChildRef::GenericParameters(self.generics()));
        children.push(ChildRef::ImplementsList(self.implements()));
        children.extend(
            self.declarations()
                .iter()
                .map(|declaration| ChildRef::InterfaceLevelDeclaration(*declaration)),
        );
        children
    }
}

impl HasChildren for ClassDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Identifier(self.identifier()));
        children.push(ChildRef::GenericParameters(self.generics()));
        if let Some(class_constructor) = self.class_constructor() {
            children.push(ChildRef::ClassConstructor(class_constructor));
        }
        children.push(ChildRef::ImplementsList(self.implements()));
        children.extend(
            self.declarations()
                .iter()
                .map(|declaration| ChildRef::ClassLevelDeclaration(*declaration)),
        );
        children
    }
}

impl HasChildren for FunctionDefinition {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Identifier(self.identifier()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
            ChildRef::FunctionBody(self.body()),
        ]
    }
}

impl HasChildren for OperatorFunctionDefinition {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
            ChildRef::FunctionBody(self.body()),
        ]
    }
}

impl HasChildren for PlatformFunctionDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Identifier(self.identifier()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
        ]
    }
}

impl HasChildren for InterfaceFunctionDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Identifier(self.identifier()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
        ];
        if let Some(body) = self.body() {
            children.push(ChildRef::FunctionBody(body))
        }
        children
    }
}

impl HasChildren for InterfaceOperatorFunctionDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![
            ChildRef::GenericParameters(self.generics()),
            ChildRef::Parameters(self.parameters()),
            ChildRef::ReturnType(self.return_type()),
        ];
        if let Some(body) = self.body() {
            children.push(ChildRef::FunctionBody(body))
        }
        children
    }
}

impl HasChildren for FunctionBody {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            Self::Equals(expression) => expression.children(),
            Self::Block(block_statement) => block_statement.children(),
            Self::Alias(identifier) => vec![ChildRef::Identifier(identifier.as_ref())],
        }
    }
}

impl HasChildren for ClassConstructor {
    fn children(&self) -> Vec<ChildRef> {
        self.parameters()
            .iter()
            .map(|parameter| ChildRef::ClassConstructorParameter(*parameter))
            .collect()
    }
}

impl HasChildren for ClassConstructorParameter {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            Self::Property(property_declaration) => property_declaration.children(),
            Self::Field(field_declaration) => field_declaration.children(),
        }
    }
}

impl HasChildren for PropertyDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Identifier(self.identifier()),
            ChildRef::TypeUse(self.declared_type()),
        ]
    }
}

impl HasChildren for FieldDeclaration {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Identifier(self.identifier()),
            ChildRef::TypeUse(self.declared_type()),
        ]
    }
}

impl HasChildren for BlockStatement {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.extend(
            self.statements()
                .iter()
                .map(|statement| ChildRef::Statement(statement)),
        );
        if let Some(expression) = self.expression() {
            children.push(ChildRef::Expression(expression));
        }
        children
    }
}

impl HasChildren for Statement {
    fn children(&self) -> Vec<ChildRef> {
        use Statement::*;
        match self {
            BlockStatement(block_statement) => block_statement.children(),
            VariableDeclarationStatement(variable_declaration_statement) => {
                variable_declaration_statement.children()
            }
            AssignStatement(assign_statement) => assign_statement.children(),
            CallStatement(call_statement) => call_statement.children(),
            ReturnStatement(return_statement) => return_statement.children(),
            IfStatement(if_statement) => if_statement.children(),
            IfElseStatement(if_else_statement) => if_else_statement.children(),
            WhileStatement(while_statement) => while_statement.children(),
            ForStatement(for_statement) => for_statement.children(),
        }
    }
}

impl HasChildren for VariableDeclarationStatement {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Identifier(self.identifier()));
        if let Some(declared_type) = self.declared_type() {
            children.push(ChildRef::TypeUse(declared_type));
        }
        if let Some(initializer) = self.initializer() {
            children.push(ChildRef::Expression(initializer));
        }
        children
    }
}

impl HasChildren for AssignStatement {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.lhs()),
            ChildRef::Expression(self.rhs()),
        ]
    }
}

impl HasChildren for CallStatement {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::Expression(self.expression())]
    }
}

impl HasChildren for ReturnStatement {
    fn children(&self) -> Vec<ChildRef> {
        if let Some(expression) = self.expression() {
            vec![ChildRef::Expression(expression)]
        } else {
            vec![]
        }
    }
}

impl HasChildren for IfStatement {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.condition()),
            ChildRef::BlockStatement(self.then_block()),
        ]
    }
}

impl HasChildren for IfElseStatement {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::IfStatement(self.if_statement()));
        children.push(ChildRef::ElseIfs(self.else_ifs()));
        if let Some(else_block) = self.else_block() {
            children.push(ChildRef::ElseBlock(else_block));
        }
        children
    }
}

impl HasChildren for ElseIfs {
    fn children(&self) -> Vec<ChildRef> {
        self.if_statements()
            .iter()
            .map(|statement| ChildRef::IfStatement(statement))
            .collect()
    }
}

impl HasChildren for ElseBlock {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::BlockStatement(self.block_statement())]
    }
}

impl HasChildren for WhileStatement {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.condition()),
            ChildRef::BlockStatement(self.body()),
        ]
    }
}

impl HasChildren for ForStatement {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Identifier(self.variable()),
            ChildRef::Expression(self.iterator()),
            ChildRef::BlockStatement(self.body()),
        ]
    }
}

impl HasChildren for Expression {
    fn children(&self) -> Vec<ChildRef> {
        use Expression::*;
        match self {
            Ternary(ternary) => ternary.children(),
            Binary(binary) => binary.children(),
            UnaryPrefix(prefix) => prefix.children(),
            UnarySuffix(suffix) => suffix.children(),
            Call(call) => call.children(),
            ObjectAccess(object_access) => object_access.children(),
            Literal(literal) => literal.children(),
            FullyQualifiedName(fully_qualified_name) => fully_qualified_name.children(),
            Closure(closure) => closure.children(),
        }
    }
}

impl HasChildren for TernaryExpression {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.condition()),
            ChildRef::Expression(self.true_expression()),
            ChildRef::Expression(self.false_expression()),
        ]
    }
}

impl HasChildren for BinaryExpression {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.left()),
            ChildRef::Expression(self.right()),
        ]
    }
}

impl HasChildren for PrefixExpression {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::Expression(self.expression())]
    }
}

impl HasChildren for SuffixExpression {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::Expression(self.expression())]
    }
}

impl HasChildren for CallExpression {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Expression(self.callee()));
        if let Some(turbo_fish) = self.turbo_fish() {
            children.push(ChildRef::TurboFish(turbo_fish));
        }
        children.push(ChildRef::CallArguments(self.arguments()));
        children
    }
}

impl HasChildren for TurboFish {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::GenericArguments(self.generics())]
    }
}

impl HasChildren for CallArguments {
    fn children(&self) -> Vec<ChildRef> {
        self.arguments()
            .iter()
            .map(|argument| ChildRef::CallArgument(*argument))
            .collect()
    }
}

impl HasChildren for CallArgument {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::Expression(self.expression())]
    }
}

impl HasChildren for Closure {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::ClosureCaptures(self.captures()));
        children.push(ChildRef::ClosureParameters(self.parameters()));
        children.extend(
            self.statements()
                .iter()
                .map(|statement| ChildRef::Statement(statement)),
        );
        if let Some(expression) = self.expression() {
            children.push(ChildRef::Expression(expression));
        }
        children
    }
}

impl HasChildren for ClosureCaptures {
    fn children(&self) -> Vec<ChildRef> {
        self.captures()
            .iter()
            .map(|capture| ChildRef::ClosureCapture(*capture))
            .collect()
    }
}

impl HasChildren for ClosureCapture {
    fn children(&self) -> Vec<ChildRef> {
        vec![ChildRef::Identifier(self.identifier())]
    }
}

impl HasChildren for ClosureParameters {
    fn children(&self) -> Vec<ChildRef> {
        self.parameters()
            .iter()
            .map(|parameter| ChildRef::ClosureParameter(*parameter))
            .collect()
    }
}

impl HasChildren for ClosureParameter {
    fn children(&self) -> Vec<ChildRef> {
        let mut children = vec![];
        children.push(ChildRef::Identifier(self.identifier()));
        if let Some(type_use) = self.type_use() {
            children.push(ChildRef::TypeUse(type_use));
        }
        children
    }
}

impl HasChildren for ObjectAccess {
    fn children(&self) -> Vec<ChildRef> {
        vec![
            ChildRef::Expression(self.receiver()),
            ChildRef::ObjectNavigations(self.navigations()),
        ]
    }
}

impl HasChildren for ObjectNavigations {
    fn children(&self) -> Vec<ChildRef> {
        self.navigations()
            .iter()
            .map(|navigation| ChildRef::ObjectNavigation(*navigation))
            .collect()
    }
}

impl HasChildren for ObjectNavigation {
    fn children(&self) -> Vec<ChildRef> {
        vec![match self {
            ObjectNavigation::Identifier(identifier) => ChildRef::Identifier(identifier),
            ObjectNavigation::Index(index_expression) => ChildRef::Expression(index_expression),
        }]
    }
}

impl HasChildren for Literal {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            Literal::DString(d_string) => vec![ChildRef::DString(d_string)],
            Literal::BacktickString(d_string) => vec![ChildRef::DString(d_string)],
            _ => vec![],
        }
    }
}

impl HasChildren for DString {
    fn children(&self) -> Vec<ChildRef> {
        self.parts()
            .iter()
            .map(|part| ChildRef::DStringPart(*part))
            .collect()
    }
}

impl HasChildren for DStringPart {
    fn children(&self) -> Vec<ChildRef> {
        match self {
            DStringPart::Expression(expression) => vec![ChildRef::Expression(expression)],
            _ => vec![],
        }
    }
}
