use crate::ast::named::Named;
use crate::name_analysis::symbol::Symbol;
use pest::Parser;
use std::borrow::Cow;
use std::range::Range;

pub mod build;
pub mod children;
pub mod named;
pub mod pretty_print;
pub mod unparse;

/* Operators */

#[derive(Debug)]
pub enum Operator {
    Binary(BinaryOperator),
    PrefixUnary(PrefixUnaryOperator),
    SuffixUnary(SuffixUnaryOperator),
}

#[derive(Debug)]
pub enum BinaryOperator {
    Or,
    And,
    EqualTo,
    NotEqualTo,
    Greater,
    Less,
    GreaterEqual,
    LessEqual,
    Add,
    Subtract,
    Multiply,
    Divide,
    Modulo,
    LeftShift,
    RightShift,
}

#[derive(Debug)]
pub enum PrefixUnaryOperator {
    Spread,
    BorrowMut,
    Borrow,
    Star,
    Mut,
    Not,
    Negative,
}

#[derive(Debug)]
pub enum SuffixUnaryOperator {
    PlusPlus,
    MinusMinus,
    Call,
    Index,
}

/* Names */

#[derive(Debug)]
pub struct Identifier {
    name: String,
    file_id: usize,
    range: Range<usize>,
    scope_id: Option<usize>,
    saved_symbol: Option<Symbol>,
}

impl Identifier {
    pub fn new(name: &str, file_id: usize, range: Range<usize>) -> Self {
        Self {
            name: name.to_string(),
            file_id,
            range,
            scope_id: None,
            saved_symbol: None,
        }
    }
}

#[derive(Debug)]
pub struct FullyQualifiedName {
    identifiers: Vec<Box<Identifier>>,
    file_id: usize,
    range: Range<usize>,
    scope_id: Option<usize>,
    saved_symbol: Option<Symbol>,
}

impl FullyQualifiedName {
    pub fn new(identifiers: Vec<Box<Identifier>>, file_id: usize, range: Range<usize>) -> Self {
        Self {
            identifiers,
            range,
            file_id,
            scope_id: None,
            saved_symbol: None,
        }
    }

    pub fn identifiers(&self) -> Vec<&Identifier> {
        self.identifiers.iter().map(Box::as_ref).collect()
    }

    pub fn identifiers_mut(&mut self) -> Vec<&mut Identifier> {
        self.identifiers.iter_mut().map(Box::as_mut).collect()
    }

    #[deprecated(note = "Grammar will instead distinguish between fqns and identifiers")]
    pub fn last(&self) -> &Identifier {
        &self.identifiers[self.identifiers.len() - 1]
    }

    #[deprecated(note = "Grammar will instead distinguish between fqns and identifiers")]
    pub fn last_mut(&mut self) -> &mut Identifier {
        let last_index = self.identifiers.len() - 1;
        &mut self.identifiers[last_index]
    }

    #[deprecated(note = "Grammar will instead distinguish between fqns and identifiers")]
    pub fn len(&self) -> usize {
        self.identifiers.len()
    }

    #[deprecated(note = "Grammar will instead distinguish between fqns and identifiers")]
    pub fn is_single_identifier(&self) -> bool {
        self.identifiers.len() == 1
    }
}

/* Type Use */

#[derive(Debug)]
pub enum TypeUse {
    Primitive(Box<PrimitiveTypeUse>),
    InterfaceOrClass(Box<InterfaceOrClassTypeUse>),
    Tuple(Box<TupleTypeUse>),
    Function(Box<FunctionTypeUse>),
}

#[derive(Debug)]
pub enum PrimitiveTypeUse {
    Byte,
    Short,
    Char,
    Int,
    Long,
    Double,
    Bool,
    String,
    Array(Option<Box<GenericArguments>>),
    Any,
    Void,
}

#[derive(Debug)]
pub struct InterfaceOrClassTypeUse {
    borrow_count: usize,
    is_mutable: bool,
    fqn: Box<FullyQualifiedName>,
    generics: Box<GenericArguments>,
}

impl InterfaceOrClassTypeUse {
    pub fn new(
        borrow_count: usize,
        is_mutable: bool,
        fqn: Box<FullyQualifiedName>,
        generics: Box<GenericArguments>,
    ) -> Self {
        Self {
            borrow_count,
            is_mutable,
            fqn,
            generics,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn fqn(&self) -> &FullyQualifiedName {
        &self.fqn
    }

    pub fn fqn_mut(&mut self) -> &mut FullyQualifiedName {
        &mut self.fqn
    }

    pub fn generics(&self) -> &GenericArguments {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericArguments {
        &mut self.generics
    }
}

#[derive(Debug)]
pub struct TupleTypeUse {
    borrow_count: usize,
    is_mutable: bool,
    arguments: Box<TupleArguments>,
}

impl TupleTypeUse {
    pub fn new(borrow_count: usize, is_mutable: bool, arguments: Box<TupleArguments>) -> Self {
        Self {
            borrow_count,
            is_mutable,
            arguments,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn arguments(&self) -> &TupleArguments {
        &self.arguments
    }

    pub fn arguments_mut(&mut self) -> &mut TupleArguments {
        &mut self.arguments
    }
}

#[derive(Debug)]
pub struct FunctionTypeUse {
    borrow_count: usize,
    function_modifier: Option<FunctionTypeModifier>,
    generics: Box<GenericParameters>,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
}

impl FunctionTypeUse {
    pub fn new(
        borrow_count: usize,
        function_modifier: Option<FunctionTypeModifier>,
        generics: Box<GenericParameters>,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
    ) -> Self {
        Self {
            borrow_count,
            function_modifier,
            generics,
            parameters,
            return_type,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn function_modifier(&self) -> Option<&FunctionTypeModifier> {
        self.function_modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn parameters(&self) -> &Box<Parameters> {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Box<Parameters> {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }
}

/* Generic arguments */

#[derive(Debug)]
pub struct GenericArguments(Vec<Box<TypeUse>>);

impl GenericArguments {
    pub fn new(arguments: Vec<Box<TypeUse>>) -> Self {
        Self(arguments)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn arguments(&self) -> Vec<&TypeUse> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn arguments_mut(&mut self) -> Vec<&mut TypeUse> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl Default for GenericArguments {
    fn default() -> Self {
        GenericArguments(Vec::new())
    }
}

/* Generic parameters */

#[derive(Debug)]
pub struct GenericParameters(Vec<Box<Identifier>>);

impl GenericParameters {
    pub fn new(identifiers: Vec<Box<Identifier>>) -> Self {
        Self(identifiers)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn identifiers(&self) -> Vec<&Identifier> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn identifiers_mut(&mut self) -> Vec<&mut Identifier> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl Default for GenericParameters {
    fn default() -> Self {
        GenericParameters(Vec::new())
    }
}

/* Tuple Arguments */

#[derive(Debug)]
pub struct TupleArguments(Vec<Box<TypeUse>>);

impl TupleArguments {
    pub fn new(arguments: Vec<Box<TypeUse>>) -> Self {
        Self(arguments)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn type_uses(&self) -> Vec<&TypeUse> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn type_uses_mut(&mut self) -> Vec<&mut TypeUse> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

/* Implements List */

#[derive(Debug)]
pub struct ImplementsList(Vec<Box<TypeUse>>);

impl ImplementsList {
    pub fn new(type_uses: Vec<Box<TypeUse>>) -> Self {
        Self(type_uses)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn type_uses(&self) -> Vec<&TypeUse> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn type_uses_mut(&mut self) -> Vec<&mut TypeUse> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl Default for ImplementsList {
    fn default() -> Self {
        ImplementsList(Vec::new())
    }
}

/* Function Type Modifier */

#[derive(Debug)]
pub enum FunctionTypeModifier {
    Cons,
    MutRef,
    Mut,
    Ref,
}

/* Function Parameters */

#[derive(Debug)]
pub struct Parameters(Vec<Box<Parameter>>);

impl Parameters {
    pub fn new(parameters: Vec<Box<Parameter>>) -> Self {
        Self(parameters)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn parameters(&self) -> Vec<&Parameter> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn parameters_mut(&mut self) -> Vec<&mut Parameter> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl Default for Parameters {
    fn default() -> Self {
        Parameters(Vec::new())
    }
}

#[derive(Debug)]
pub struct Parameter {
    identifier: Box<Identifier>,
    type_use: Box<TypeUse>,
}

impl Parameter {
    pub fn new(identifier: Box<Identifier>, type_use: Box<TypeUse>) -> Self {
        Self {
            identifier,
            type_use,
        }
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn type_use(&self) -> &TypeUse {
        &self.type_use
    }

    pub fn type_use_mut(&mut self) -> &mut TypeUse {
        &mut self.type_use
    }
}

/* Return Type */

#[derive(Debug)]
pub struct ReturnType {
    declared_type: Box<TypeUse>,
    references: Box<References>,
}

impl ReturnType {
    pub fn new(declared_type: Box<TypeUse>, references: Box<References>) -> Self {
        Self {
            declared_type,
            references,
        }
    }

    pub fn void() -> Self {
        ReturnType {
            declared_type: Box::new(TypeUse::Primitive(Box::new(PrimitiveTypeUse::Void))),
            references: Box::new(References::default()),
        }
    }

    pub fn declared_type(&self) -> &TypeUse {
        &self.declared_type
    }

    pub fn declared_type_mut(&mut self) -> &mut TypeUse {
        &mut self.declared_type
    }

    pub fn references(&self) -> &References {
        &self.references
    }

    pub fn references_mut(&mut self) -> &mut References {
        &mut self.references
    }
}

/* References */

#[derive(Debug)]
pub struct References(Vec<Box<Identifier>>);

impl References {
    pub fn new(identifiers: Vec<Box<Identifier>>) -> Self {
        Self(identifiers)
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn identifiers(&self) -> Vec<&Identifier> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn identifiers_mut(&mut self) -> Vec<&mut Identifier> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl Default for References {
    fn default() -> Self {
        References(Vec::new())
    }
}

/* Top-level construct */

#[derive(Debug)]
pub struct CompilationUnit {
    file_name: String,
    file_id: usize,
    namespace: Option<Box<FullyQualifiedName>>,
    use_statements: Vec<Box<UseStatement>>,
    declarations: Vec<Box<ModuleLevelDeclaration>>,
}

impl CompilationUnit {
    pub fn new(
        file_name: String,
        file_id: usize,
        namespace: Option<Box<FullyQualifiedName>>,
        use_statements: Vec<Box<UseStatement>>,
        declarations: Vec<Box<ModuleLevelDeclaration>>,
    ) -> Self {
        Self {
            file_name,
            file_id,
            namespace,
            use_statements,
            declarations,
        }
    }

    pub fn file_name(&self) -> &str {
        &self.file_name
    }

    pub fn file_id(&self) -> usize {
        self.file_id
    }

    pub fn namespace(&self) -> Option<&FullyQualifiedName> {
        // We have to use this non-map syntax because we need a reference like `&self.namespace`
        if let Some(namespace) = &self.namespace {
            Some(namespace.as_ref())
        } else {
            None
        }
    }

    pub fn use_statements(&self) -> Vec<&UseStatement> {
        self.use_statements.iter().map(Box::as_ref).collect()
    }

    pub fn use_statements_mut(&mut self) -> Vec<&mut UseStatement> {
        self.use_statements.iter_mut().map(Box::as_mut).collect()
    }

    pub fn declarations(&self) -> Vec<&ModuleLevelDeclaration> {
        self.declarations.iter().map(Box::as_ref).collect()
    }

    pub fn declarations_mut(&mut self) -> Vec<&mut ModuleLevelDeclaration> {
        self.declarations.iter_mut().map(Box::as_mut).collect()
    }
}

/* Use Statement */

#[derive(Debug)]
pub struct UseStatement {
    identifiers: Vec<Box<Identifier>>,
    last: Box<UseStatementLast>,
    file_id: usize,
    range: Range<usize>,
}

impl UseStatement {
    pub fn new(
        identifiers: Vec<Box<Identifier>>,
        last: Box<UseStatementLast>,
        file_id: usize,
        range: Range<usize>,
    ) -> Self {
        UseStatement {
            identifiers,
            last,
            file_id,
            range,
        }
    }

    pub fn base_name(&self) -> Cow<'_, str> {
        use UseStatementLast::*;
        if self.identifiers.is_empty() {
            match self.last.as_ref() {
                Identifier(_) => Cow::from(""),
                Star | Identifiers(_) => panic!(), // should never get here because of grammar
            }
        } else if self.identifiers.len() == 1 {
            self.identifiers[0].name()
        } else {
            let mut acc = String::new();
            for (i, identifier) in self.identifiers.iter().enumerate() {
                acc.push_str(&identifier.name());
                if i != self.identifiers.len() - 1 {
                    acc.push_str("::");
                }
            }
            Cow::from(acc)
        }
    }

    pub fn is_star(&self) -> bool {
        match self.last.as_ref() {
            UseStatementLast::Star => true,
            _ => false,
        }
    }

    pub fn identifiers(&self) -> Vec<&Identifier> {
        self.identifiers.iter().map(Box::as_ref).collect()
    }

    pub fn identifiers_mut(&mut self) -> Vec<&mut Identifier> {
        self.identifiers.iter_mut().map(Box::as_mut).collect()
    }

    pub fn last(&self) -> &UseStatementLast {
        &self.last
    }

    pub fn last_mut(&mut self) -> &mut UseStatementLast {
        &mut self.last
    }

    pub fn file_id(&self) -> usize {
        self.file_id
    }

    pub fn range(&self) -> Range<usize> {
        self.range
    }
}

#[derive(Debug)]
pub enum UseStatementLast {
    Identifier(Box<Identifier>),
    Identifiers(Vec<Box<Identifier>>),
    Star,
}

/* Declarations allowed in each level */

#[derive(Debug)]
pub enum ModuleLevelDeclaration {
    Module(Box<ModuleDeclaration>),
    Interface(Box<InterfaceDeclaration>),
    Class(Box<ClassDeclaration>),
    Function(Box<FunctionDefinition>),
    PlatformFunction(Box<PlatformFunctionDeclaration>),
}

#[derive(Debug)]
pub enum InterfaceLevelDeclaration {
    Module(Box<ModuleDeclaration>),
    Interface(Box<InterfaceDeclaration>),
    Class(Box<ClassDeclaration>),
    Function(Box<InterfaceFunctionDeclaration>),
    OperatorFunction(Box<InterfaceOperatorFunctionDeclaration>),
}

#[derive(Debug)]
pub enum ClassLevelDeclaration {
    Module(Box<ModuleDeclaration>),
    Interface(Box<InterfaceDeclaration>),
    Class(Box<ClassDeclaration>),
    Function(Box<FunctionDefinition>),
    OperatorFunction(Box<OperatorFunctionDefinition>),
    PlatformFunction(Box<PlatformFunctionDeclaration>),
    Property(Box<PropertyDeclaration>),
    Field(Box<FieldDeclaration>),
}

/* Main Declarations */

#[derive(Debug)]
pub struct ModuleDeclaration {
    is_public: bool,
    identifier: Box<Identifier>,
    declarations: Vec<Box<ModuleLevelDeclaration>>,
}

impl ModuleDeclaration {
    pub fn new(
        is_public: bool,
        identifier: Box<Identifier>,
        declarations: Vec<Box<ModuleLevelDeclaration>>,
    ) -> Self {
        Self {
            is_public,
            identifier,
            declarations,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declarations(&self) -> Vec<&ModuleLevelDeclaration> {
        self.declarations.iter().map(Box::as_ref).collect()
    }

    pub fn declarations_mut(&mut self) -> Vec<&mut ModuleLevelDeclaration> {
        self.declarations.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub struct InterfaceDeclaration {
    is_public: bool,
    identifier: Box<Identifier>,
    generics: Box<GenericParameters>,
    implements: Box<ImplementsList>,
    declarations: Vec<Box<InterfaceLevelDeclaration>>,
}

impl InterfaceDeclaration {
    pub fn new(
        is_public: bool,
        identifier: Box<Identifier>,
        generics: Box<GenericParameters>,
        implements: Box<ImplementsList>,
        declarations: Vec<Box<InterfaceLevelDeclaration>>,
    ) -> Self {
        Self {
            is_public,
            identifier,
            generics,
            implements,
            declarations,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn implements(&self) -> &ImplementsList {
        &self.implements
    }

    pub fn implements_mut(&mut self) -> &mut ImplementsList {
        &mut self.implements
    }

    pub fn declarations(&self) -> Vec<&InterfaceLevelDeclaration> {
        self.declarations.iter().map(Box::as_ref).collect()
    }

    pub fn declarations_mut(&mut self) -> Vec<&mut InterfaceLevelDeclaration> {
        self.declarations.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub struct ClassDeclaration {
    is_public: bool,
    identifier: Box<Identifier>,
    generics: Box<GenericParameters>,
    class_constructor: Option<Box<ClassConstructor>>,
    implements: Box<ImplementsList>,
    declarations: Vec<Box<ClassLevelDeclaration>>,
}

impl ClassDeclaration {
    pub fn new(
        is_public: bool,
        identifier: Box<Identifier>,
        generics: Box<GenericParameters>,
        class_constructor: Option<Box<ClassConstructor>>,
        implements: Box<ImplementsList>,
        declarations: Vec<Box<ClassLevelDeclaration>>,
    ) -> Self {
        Self {
            is_public,
            identifier,
            generics,
            class_constructor,
            implements,
            declarations,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn class_constructor(&self) -> Option<&ClassConstructor> {
        if let Some(class_constructor) = &self.class_constructor {
            Some(class_constructor.as_ref())
        } else {
            None
        }
    }

    pub fn class_constructor_mut(&mut self) -> Option<&mut ClassConstructor> {
        if let Some(class_constructor) = &mut self.class_constructor {
            Some(class_constructor.as_mut())
        } else {
            None
        }
    }

    pub fn implements(&self) -> &ImplementsList {
        &self.implements
    }

    pub fn implements_mut(&mut self) -> &mut ImplementsList {
        &mut self.implements
    }

    pub fn declarations(&self) -> Vec<&ClassLevelDeclaration> {
        self.declarations.iter().map(Box::as_ref).collect()
    }

    pub fn declarations_mut(&mut self) -> Vec<&mut ClassLevelDeclaration> {
        self.declarations.iter_mut().map(Box::as_mut).collect()
    }
}

/* Function declarations and components */

#[derive(Debug)]
pub struct FunctionDefinition {
    is_public: bool,
    modifier: Option<FunctionModifier>,
    generics: Box<GenericParameters>,
    identifier: Box<Identifier>,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
    body: Box<FunctionBody>,
}

impl FunctionDefinition {
    pub fn new(
        is_public: bool,
        modifier: Option<FunctionModifier>,
        generics: Box<GenericParameters>,
        identifier: Box<Identifier>,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
        body: Box<FunctionBody>,
    ) -> Self {
        Self {
            is_public,
            modifier,
            generics,
            identifier,
            parameters,
            return_type,
            body,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn modifier(&self) -> Option<&FunctionModifier> {
        self.modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn parameters(&self) -> &Parameters {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Parameters {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }

    pub fn body(&self) -> &FunctionBody {
        &self.body
    }

    pub fn body_mut(&mut self) -> &mut FunctionBody {
        &mut self.body
    }
}

#[derive(Debug)]
pub struct OperatorFunctionDefinition {
    is_public: bool,
    modifier: Option<FunctionModifier>,
    generics: Box<GenericParameters>,
    operator: Operator,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
    body: Box<FunctionBody>,
}

impl OperatorFunctionDefinition {
    pub fn new(
        is_public: bool,
        modifier: Option<FunctionModifier>,
        generics: Box<GenericParameters>,
        operator: Operator,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
        body: Box<FunctionBody>,
    ) -> Self {
        Self {
            is_public,
            modifier,
            generics,
            operator,
            parameters,
            return_type,
            body,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn modifier(&self) -> Option<&FunctionModifier> {
        self.modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn operator(&self) -> &Operator {
        &self.operator
    }

    pub fn parameters(&self) -> &Parameters {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Parameters {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }

    pub fn body(&self) -> &FunctionBody {
        &self.body
    }

    pub fn body_mut(&mut self) -> &mut FunctionBody {
        &mut self.body
    }
}

#[derive(Debug)]
pub struct PlatformFunctionDeclaration {
    is_public: bool,
    modifier: Option<FunctionModifier>,
    generics: Box<GenericParameters>,
    identifier: Box<Identifier>,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
}

impl PlatformFunctionDeclaration {
    pub fn new(
        is_public: bool,
        modifier: Option<FunctionModifier>,
        generics: Box<GenericParameters>,
        identifier: Box<Identifier>,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
    ) -> Self {
        Self {
            is_public,
            modifier,
            generics,
            identifier,
            parameters,
            return_type,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn modifier(&self) -> Option<&FunctionModifier> {
        self.modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn parameters(&self) -> &Parameters {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Parameters {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }
}

#[derive(Debug)]
pub struct InterfaceFunctionDeclaration {
    modifier: Option<FunctionModifier>,
    generics: Box<GenericParameters>,
    identifier: Box<Identifier>,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
    body: Option<Box<FunctionBody>>,
}

impl InterfaceFunctionDeclaration {
    pub fn new(
        modifier: Option<FunctionModifier>,
        generics: Box<GenericParameters>,
        identifier: Box<Identifier>,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
        body: Option<Box<FunctionBody>>,
    ) -> Self {
        Self {
            modifier,
            generics,
            identifier,
            parameters,
            return_type,
            body,
        }
    }

    pub fn modifier(&self) -> Option<&FunctionModifier> {
        self.modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn parameters(&self) -> &Parameters {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Parameters {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }

    pub fn body(&self) -> Option<&FunctionBody> {
        FunctionBody::unwrap_option(&self.body)
    }

    pub fn body_mut(&mut self) -> Option<&mut FunctionBody> {
        FunctionBody::unwrap_option_mut(&mut self.body)
    }
}

#[derive(Debug)]
pub struct InterfaceOperatorFunctionDeclaration {
    modifier: Option<FunctionModifier>,
    generics: Box<GenericParameters>,
    operator: Operator,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
    body: Option<Box<FunctionBody>>,
}

impl InterfaceOperatorFunctionDeclaration {
    pub fn new(
        modifier: Option<FunctionModifier>,
        generics: Box<GenericParameters>,
        operator: Operator,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
        body: Option<Box<FunctionBody>>,
    ) -> Self {
        Self {
            modifier,
            generics,
            operator,
            parameters,
            return_type,
            body,
        }
    }

    pub fn modifier(&self) -> Option<&FunctionModifier> {
        self.modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn operator(&self) -> &Operator {
        &self.operator
    }

    pub fn parameters(&self) -> &Box<Parameters> {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Box<Parameters> {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }

    pub fn body(&self) -> Option<&FunctionBody> {
        FunctionBody::unwrap_option(&self.body)
    }

    pub fn body_mut(&mut self) -> Option<&mut FunctionBody> {
        FunctionBody::unwrap_option_mut(&mut self.body)
    }
}

#[derive(Debug)]
pub enum FunctionModifier {
    Static,
    Cons,
    Mut,
    Ref,
    MutRef,
}

#[derive(Debug)]
pub enum FunctionBody {
    Equals(Box<Expression>),
    Block(Box<BlockStatement>),
    Alias(Box<Identifier>),
}

impl FunctionBody {
    fn unwrap_option(opt: &Option<Box<FunctionBody>>) -> Option<&FunctionBody> {
        if let Some(body) = opt {
            Some(body.as_ref())
        } else {
            None
        }
    }

    fn unwrap_option_mut(opt: &mut Option<Box<FunctionBody>>) -> Option<&mut FunctionBody> {
        if let Some(body) = opt {
            Some(body.as_mut())
        } else {
            None
        }
    }
}

/* Class components */

#[derive(Debug)]
pub struct ClassConstructor(Vec<Box<ClassConstructorParameter>>);

impl ClassConstructor {
    pub fn new(parameters: Vec<Box<ClassConstructorParameter>>) -> Self {
        Self(parameters)
    }

    pub fn parameters(&self) -> Vec<&ClassConstructorParameter> {
        self.0.iter().map(|p| p.as_ref()).collect()
    }
    
    pub fn parameters_mut(&mut self) -> Vec<&mut ClassConstructorParameter> {
        self.0.iter_mut().map(|p| p.as_mut()).collect()
    }
}

#[derive(Debug)]
pub enum ClassConstructorParameter {
    Property(Box<PropertyDeclaration>),
    Field(Box<FieldDeclaration>),
}

#[derive(Debug)]
pub struct PropertyDeclaration {
    is_mutable: bool,
    identifier: Box<Identifier>,
    declared_type: Box<TypeUse>,
}

impl PropertyDeclaration {
    pub fn new(is_mutable: bool, identifier: Box<Identifier>, declared_type: Box<TypeUse>) -> Self {
        Self {
            is_mutable,
            identifier,
            declared_type,
        }
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declared_type(&self) -> &TypeUse {
        &self.declared_type
    }

    pub fn declared_type_mut(&mut self) -> &mut TypeUse {
        &mut self.declared_type
    }
}

#[derive(Debug)]
pub struct FieldDeclaration {
    is_mutable: bool,
    identifier: Box<Identifier>,
    declared_type: Box<TypeUse>,
}

impl FieldDeclaration {
    pub fn new(is_mutable: bool, identifier: Box<Identifier>, declared_type: Box<TypeUse>) -> Self {
        Self {
            is_mutable,
            identifier,
            declared_type,
        }
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declared_type(&self) -> &TypeUse {
        &self.declared_type
    }

    pub fn declared_type_mut(&mut self) -> &mut TypeUse {
        &mut self.declared_type
    }
}

/* Statements */

#[derive(Debug)]
pub struct BlockStatement {
    statements: Vec<Box<Statement>>,
    expression: Option<Box<Expression>>,
}

impl BlockStatement {
    pub fn new(statements: Vec<Box<Statement>>, expression: Option<Box<Expression>>) -> Self {
        Self {
            statements,
            expression,
        }
    }

    pub fn statements(&self) -> Vec<&Statement> {
        self.statements.iter().map(AsRef::as_ref).collect()
    }

    pub fn statements_mut(&mut self) -> Vec<&mut Statement> {
        self.statements.iter_mut().map(AsMut::as_mut).collect()
    }

    pub fn expression(&self) -> Option<&Expression> {
        if let Some(ref expression) = self.expression {
            Some(expression.as_ref())
        } else {
            None
        }
    }

    pub fn expression_mut(&mut self) -> Option<&mut Expression> {
        if let Some(ref mut expression) = self.expression {
            Some(expression.as_mut())
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub enum Statement {
    BlockStatement(Box<BlockStatement>),
    VariableDeclarationStatement(Box<VariableDeclarationStatement>),
    AssignStatement(Box<AssignStatement>),
    CallStatement(Box<CallStatement>),
    ReturnStatement(Box<ReturnStatement>),
    IfStatement(Box<IfStatement>),
    IfElseStatement(Box<IfElseStatement>),
    WhileStatement(Box<WhileStatement>),
    ForStatement(Box<ForStatement>),
}

#[derive(Debug)]
pub struct VariableDeclarationStatement {
    is_mutable: bool,
    identifier: Box<Identifier>,
    declared_type: Option<Box<TypeUse>>,
    initializer: Option<Box<Expression>>,
}

impl VariableDeclarationStatement {
    pub fn new(
        is_mutable: bool,
        identifier: Box<Identifier>,
        declared_type: Option<Box<TypeUse>>,
        initializer: Option<Box<Expression>>,
    ) -> Self {
        Self {
            is_mutable,
            identifier,
            declared_type,
            initializer,
        }
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declared_type(&self) -> Option<&TypeUse> {
        if let Some(type_use) = &self.declared_type {
            Some(type_use.as_ref())
        } else {
            None
        }
    }

    pub fn declared_type_mut(&mut self) -> Option<&mut TypeUse> {
        if let Some(type_use) = &mut self.declared_type {
            Some(type_use.as_mut())
        } else {
            None
        }
    }

    pub fn initializer(&self) -> Option<&Expression> {
        if let Some(initializer) = &self.initializer {
            Some(initializer.as_ref())
        } else {
            None
        }
    }

    pub fn initializer_mut(&mut self) -> Option<&mut Expression> {
        if let Some(initializer) = &mut self.initializer {
            Some(initializer.as_mut())
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub struct AssignStatement {
    lhs: Box<Expression>,
    rhs: Box<Expression>,
}

impl AssignStatement {
    pub fn new(lhs: Box<Expression>, rhs: Box<Expression>) -> Self {
        Self { lhs, rhs }
    }

    pub fn lhs(&self) -> &Expression {
        &self.lhs
    }

    pub fn lhs_mut(&mut self) -> &mut Expression {
        &mut self.lhs
    }

    pub fn rhs(&self) -> &Expression {
        &self.rhs
    }

    pub fn rhs_mut(&mut self) -> &mut Expression {
        &mut self.rhs
    }
}

#[derive(Debug)]
pub struct CallStatement(Box<Expression>);

impl CallStatement {
    pub fn new(call_expression: Box<Expression>) -> Self {
        Self(call_expression)
    }

    pub fn expression(&self) -> &Expression {
        &self.0
    }

    pub fn expression_mut(&mut self) -> &mut Expression {
        &mut self.0
    }
}

#[derive(Debug)]
pub struct ReturnStatement(Option<Box<Expression>>);

impl ReturnStatement {
    pub fn new(expression: Option<Box<Expression>>) -> Self {
        Self(expression)
    }

    pub fn expression(&self) -> Option<&Expression> {
        if let Some(expression) = &self.0 {
            Some(expression.as_ref())
        } else {
            None
        }
    }

    pub fn expression_mut(&mut self) -> Option<&mut Expression> {
        if let Some(expression) = &mut self.0 {
            Some(expression.as_mut())
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub struct IfStatement {
    condition: Box<Expression>,
    then_block: Box<BlockStatement>,
}

impl IfStatement {
    pub fn new(condition: Box<Expression>, then_block: Box<BlockStatement>) -> Self {
        Self {
            condition,
            then_block,
        }
    }

    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    pub fn condition_mut(&mut self) -> &mut Expression {
        &mut self.condition
    }

    pub fn then_block(&self) -> &BlockStatement {
        &self.then_block
    }

    pub fn then_block_mut(&mut self) -> &mut BlockStatement {
        &mut self.then_block
    }
}

#[derive(Debug)]
pub struct IfElseStatement {
    if_statement: Box<IfStatement>,
    else_ifs: Box<ElseIfs>,
    else_block: Option<Box<ElseBlock>>,
}

impl IfElseStatement {
    pub fn new(
        if_statement: Box<IfStatement>,
        else_ifs: Box<ElseIfs>,
        else_block: Option<Box<ElseBlock>>,
    ) -> Self {
        Self {
            if_statement,
            else_ifs,
            else_block,
        }
    }

    pub fn if_statement(&self) -> &IfStatement {
        &self.if_statement
    }

    pub fn if_statement_mut(&mut self) -> &mut IfStatement {
        &mut self.if_statement
    }

    pub fn else_ifs(&self) -> &ElseIfs {
        &self.else_ifs
    }

    pub fn else_ifs_mut(&mut self) -> &mut ElseIfs {
        &mut self.else_ifs
    }

    pub fn else_block(&self) -> Option<&ElseBlock> {
        if let Some(else_block) = &self.else_block {
            Some(else_block.as_ref())
        } else {
            None
        }
    }

    pub fn else_block_mut(&mut self) -> Option<&mut ElseBlock> {
        if let Some(else_block) = &mut self.else_block {
            Some(else_block.as_mut())
        } else {
            None
        }
    }
}

#[derive(Debug, Default)]
pub struct ElseIfs(Vec<Box<IfStatement>>);

impl ElseIfs {
    pub fn new(if_statements: Vec<Box<IfStatement>>) -> Self {
        Self(if_statements)
    }

    pub fn if_statements(&self) -> Vec<&IfStatement> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn if_statements_mut(&mut self) -> Vec<&mut IfStatement> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub struct ElseBlock(Box<BlockStatement>);

impl ElseBlock {
    pub fn new(block: Box<BlockStatement>) -> Self {
        Self(block)
    }

    pub fn block_statement(&self) -> &BlockStatement {
        &self.0
    }

    pub fn block_statement_mut(&mut self) -> &mut BlockStatement {
        &mut self.0
    }
}

#[derive(Debug)]
pub struct WhileStatement {
    condition: Box<Expression>,
    body: Box<BlockStatement>,
}

impl WhileStatement {
    pub fn new(condition: Box<Expression>, body: Box<BlockStatement>) -> Self {
        Self { condition, body }
    }

    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    pub fn condition_mut(&mut self) -> &mut Expression {
        &mut self.condition
    }

    pub fn body(&self) -> &BlockStatement {
        &self.body
    }

    pub fn body_mut(&mut self) -> &mut BlockStatement {
        &mut self.body
    }
}

#[derive(Debug)]
pub struct ForStatement {
    variable: Box<Identifier>,
    iterator: Box<Expression>,
    body: Box<BlockStatement>,
}

impl ForStatement {
    pub fn new(
        variable: Box<Identifier>,
        iterator: Box<Expression>,
        body: Box<BlockStatement>,
    ) -> Self {
        Self {
            variable,
            iterator,
            body,
        }
    }

    pub fn variable(&self) -> &Identifier {
        &self.variable
    }

    pub fn variable_mut(&mut self) -> &mut Identifier {
        &mut self.variable
    }

    pub fn iterator(&self) -> &Expression {
        &self.iterator
    }

    pub fn iterator_mut(&mut self) -> &mut Expression {
        &mut self.iterator
    }

    pub fn body(&self) -> &BlockStatement {
        &self.body
    }

    pub fn body_mut(&mut self) -> &mut BlockStatement {
        &mut self.body
    }
}

/* Expressions */

#[derive(Debug)]
pub enum Expression {
    Ternary(Box<TernaryExpression>),
    Binary(Box<BinaryExpression>),
    UnaryPrefix(Box<PrefixExpression>),
    UnarySuffix(Box<SuffixExpression>),
    Call(Box<CallExpression>),
    ObjectAccess(Box<ObjectAccess>),
    Literal(Box<Literal>),
    FullyQualifiedName(Box<FullyQualifiedName>),
    Closure(Box<Closure>),
}

#[derive(Debug)]
pub struct TernaryExpression {
    condition: Box<Expression>,
    true_expression: Box<Expression>,
    false_expression: Box<Expression>,
}

impl TernaryExpression {
    pub fn new(
        condition: Box<Expression>,
        true_expression: Box<Expression>,
        false_expression: Box<Expression>,
    ) -> Self {
        Self {
            condition,
            true_expression,
            false_expression,
        }
    }

    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    pub fn condition_mut(&mut self) -> &mut Expression {
        &mut self.condition
    }

    pub fn true_expression(&self) -> &Expression {
        &self.true_expression
    }

    pub fn true_expression_mut(&mut self) -> &mut Expression {
        &mut self.true_expression
    }

    pub fn false_expression(&self) -> &Expression {
        &self.false_expression
    }

    pub fn false_expression_mut(&mut self) -> &mut Expression {
        &mut self.false_expression
    }
}

#[derive(Debug)]
pub struct BinaryExpression {
    left: Box<Expression>,
    operator: BinaryOperator,
    right: Box<Expression>,
}

impl BinaryExpression {
    pub fn new(left: Box<Expression>, operator: BinaryOperator, right: Box<Expression>) -> Self {
        Self {
            left,
            operator,
            right,
        }
    }

    pub fn left(&self) -> &Expression {
        &self.left
    }

    pub fn left_mut(&mut self) -> &mut Expression {
        &mut self.left
    }

    pub fn operator(&self) -> &BinaryOperator {
        &self.operator
    }

    pub fn right(&self) -> &Expression {
        &self.right
    }

    pub fn right_mut(&mut self) -> &mut Expression {
        &mut self.right
    }
}

#[derive(Debug)]
pub struct PrefixExpression {
    operator: PrefixUnaryOperator,
    expression: Box<Expression>,
}

impl PrefixExpression {
    pub fn new(operator: PrefixUnaryOperator, expression: Box<Expression>) -> Self {
        Self {
            operator,
            expression,
        }
    }

    pub fn operator(&self) -> &PrefixUnaryOperator {
        &self.operator
    }

    pub fn expression(&self) -> &Expression {
        &self.expression
    }

    pub fn expression_mut(&mut self) -> &mut Expression {
        &mut self.expression
    }
}

#[derive(Debug)]
pub struct SuffixExpression {
    expression: Box<Expression>,
    operator: SuffixUnaryOperator,
}

impl SuffixExpression {
    pub fn new(expression: Box<Expression>, operator: SuffixUnaryOperator) -> Self {
        Self {
            expression,
            operator,
        }
    }

    pub fn expression(&self) -> &Expression {
        &self.expression
    }

    pub fn expression_mut(&mut self) -> &mut Expression {
        &mut self.expression
    }

    pub fn operator(&self) -> &SuffixUnaryOperator {
        &self.operator
    }
}

#[derive(Debug)]
pub struct CallExpression {
    callee: Box<Expression>,
    turbo_fish: Option<Box<TurboFish>>,
    arguments: Box<CallArguments>,
}

impl CallExpression {
    pub fn new(
        callee: Box<Expression>,
        turbo_fish: Option<Box<TurboFish>>,
        arguments: Box<CallArguments>,
    ) -> Self {
        Self {
            callee,
            turbo_fish,
            arguments,
        }
    }

    pub fn callee(&self) -> &Expression {
        &self.callee
    }

    pub fn callee_mut(&mut self) -> &mut Expression {
        &mut self.callee
    }

    pub fn turbo_fish(&self) -> Option<&TurboFish> {
        if let Some(turbo_fish) = &self.turbo_fish {
            Some(turbo_fish.as_ref())
        } else {
            None
        }
    }

    pub fn turbo_fish_mut(&mut self) -> Option<&mut TurboFish> {
        if let Some(turbo_fish) = &mut self.turbo_fish {
            Some(turbo_fish.as_mut())
        } else {
            None
        }
    }

    pub fn arguments(&self) -> &CallArguments {
        &self.arguments
    }

    pub fn arguments_mut(&mut self) -> &mut CallArguments {
        &mut self.arguments
    }
}

#[derive(Debug)]
pub struct TurboFish(Box<GenericArguments>);

impl TurboFish {
    pub fn new(generics: Box<GenericArguments>) -> Self {
        Self(generics)
    }

    pub fn generics(&self) -> &GenericArguments {
        &self.0
    }

    pub fn generics_mut(&mut self) -> &mut GenericArguments {
        &mut self.0
    }
}

#[derive(Debug)]
pub struct CallArguments(Vec<Box<CallArgument>>);

impl CallArguments {
    pub fn new(arguments: Vec<Box<CallArgument>>) -> Self {
        Self(arguments)
    }

    pub fn arguments(&self) -> Vec<&CallArgument> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn arguments_mut(&mut self) -> Vec<&mut CallArgument> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub struct CallArgument(Box<Expression>);

impl CallArgument {
    pub fn new(expression: Box<Expression>) -> Self {
        Self(expression)
    }

    pub fn expression(&self) -> &Expression {
        &self.0
    }

    pub fn expression_mut(&mut self) -> &mut Expression {
        &mut self.0
    }
}

#[derive(Debug)]
pub struct Closure {
    modifier: Option<ClosureModifier>,
    is_move: bool,
    captures: Box<ClosureCaptures>,
    parameters: Box<ClosureParameters>,
    statements: Vec<Box<Statement>>,
    expression: Option<Box<Expression>>,
}

impl Closure {
    pub fn new(
        modifier: Option<ClosureModifier>,
        is_move: bool,
        captures: Box<ClosureCaptures>,
        parameters: Box<ClosureParameters>,
        statements: Vec<Box<Statement>>,
        expression: Option<Box<Expression>>,
    ) -> Self {
        Self {
            modifier,
            is_move,
            captures,
            parameters,
            statements,
            expression,
        }
    }

    pub fn modifier(&self) -> Option<&ClosureModifier> {
        self.modifier.as_ref()
    }

    pub fn is_move(&self) -> bool {
        self.is_move
    }

    pub fn captures(&self) -> &ClosureCaptures {
        &self.captures
    }

    pub fn captures_mut(&mut self) -> &mut ClosureCaptures {
        &mut self.captures
    }

    pub fn parameters(&self) -> &ClosureParameters {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut ClosureParameters {
        &mut self.parameters
    }

    pub fn statements(&self) -> Vec<&Statement> {
        self.statements.iter().map(Box::as_ref).collect()
    }

    pub fn statements_mut(&mut self) -> Vec<&mut Statement> {
        self.statements.iter_mut().map(Box::as_mut).collect()
    }

    pub fn expression(&self) -> Option<&Expression> {
        if let Some(expression) = &self.expression {
            Some(expression)
        } else {
            None
        }
    }

    pub fn expression_mut(&mut self) -> Option<&mut Expression> {
        if let Some(expression) = &mut self.expression {
            Some(expression)
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub enum ClosureModifier {
    Cons,
    Mut,
}

#[derive(Debug, Default)]
pub struct ClosureCaptures(Vec<Box<ClosureCapture>>);

impl ClosureCaptures {
    pub fn new(captures: Vec<Box<ClosureCapture>>) -> Self {
        Self(captures)
    }

    pub fn captures(&self) -> Vec<&ClosureCapture> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn captures_mut(&mut self) -> Vec<&mut ClosureCapture> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

impl ClosureCaptures {
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct ClosureCapture {
    borrow_count: usize,
    is_mutable: bool,
    identifier: Box<Identifier>,
}

impl ClosureCapture {
    pub fn new(borrow_count: usize, is_mutable: bool, identifier: Box<Identifier>) -> Self {
        Self {
            borrow_count,
            is_mutable,
            identifier,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }
}

#[derive(Debug, Default)]
pub struct ClosureParameters(Vec<Box<ClosureParameter>>);

impl ClosureParameters {
    pub fn new(parameters: Vec<Box<ClosureParameter>>) -> Self {
        Self(parameters)
    }

    pub fn parameters(&self) -> Vec<&ClosureParameter> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn parameters_mut(&mut self) -> Vec<&mut ClosureParameter> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

#[derive(Debug)]
pub struct ClosureParameter {
    identifier: Box<Identifier>,
    type_use: Option<Box<TypeUse>>,
}

impl ClosureParameter {
    pub fn new(identifier: Box<Identifier>, type_use: Option<Box<TypeUse>>) -> Self {
        Self {
            identifier,
            type_use,
        }
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn type_use(&self) -> Option<&TypeUse> {
        if let Some(type_use) = &self.type_use {
            Some(type_use)
        } else {
            None
        }
    }

    pub fn type_use_mut(&mut self) -> Option<&mut TypeUse> {
        if let Some(type_use) = &mut self.type_use {
            Some(type_use.as_mut())
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub struct ObjectAccess {
    receiver: Box<Expression>,
    navigations: Box<ObjectNavigations>,
}

impl ObjectAccess {
    pub fn new(receiver: Box<Expression>, navigations: Box<ObjectNavigations>) -> Self {
        Self {
            receiver,
            navigations,
        }
    }

    pub fn receiver(&self) -> &Expression {
        &self.receiver
    }

    pub fn receiver_mut(&mut self) -> &mut Expression {
        &mut self.receiver
    }

    pub fn navigations(&self) -> &ObjectNavigations {
        &self.navigations
    }

    pub fn navigations_mut(&mut self) -> &mut ObjectNavigations {
        &mut self.navigations
    }
}

#[derive(Debug)]
pub struct ObjectNavigations(Vec<Box<ObjectNavigation>>);

impl ObjectNavigations {
    pub fn new(navigations: Vec<Box<ObjectNavigation>>) -> Self {
        Self(navigations)
    }

    pub fn navigations(&self) -> Vec<&ObjectNavigation> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn navigations_mut(&mut self) -> Vec<&mut ObjectNavigation> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub enum ObjectNavigation {
    Index(Box<Expression>),
    Identifier(Box<Identifier>),
}

#[derive(Debug)]
pub enum Literal {
    Integer(i32),
    Long(i64),
    Double(f64),
    USize(usize),
    String(String),
    DString(Box<DString>),
    BacktickString(Box<DString>),
    Boolean(bool),
}

#[derive(Debug)]
pub struct DString(Vec<Box<DStringPart>>);

impl DString {
    pub fn new(parts: Vec<Box<DStringPart>>) -> Self {
        Self(parts)
    }

    pub fn parts(&self) -> Vec<&DStringPart> {
        self.0.iter().map(Box::as_ref).collect()
    }

    pub fn parts_mut(&mut self) -> Vec<&mut DStringPart> {
        self.0.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub enum DStringPart {
    String(String),
    Expression(Box<Expression>),
}

impl DStringPart {
    pub fn from_string(s: &str) -> DStringPart {
        DStringPart::String(String::from(s))
    }

    pub fn from_expression(e: Expression) -> DStringPart {
        DStringPart::Expression(Box::new(e))
    }

    pub fn is_string(&self) -> bool {
        match self {
            DStringPart::String(_) => true,
            _ => false,
        }
    }

    pub fn unwrap_string(self) -> String {
        match self {
            DStringPart::String(s) => s,
            _ => panic!(),
        }
    }

    pub fn is_expression(&self) -> bool {
        match self {
            DStringPart::Expression(_) => true,
            _ => false,
        }
    }
}
